/*
 *  lmEQTL.c
 *
 *  Created by Wei Sun on 10/08/2007.
 *
 */

#include <stdio.h>
#include <stddef.h>
#include <time.h>
#include <string.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_statistics_double.h>
#include <R.h>
#include "utility.h"

/**********************************************************************
 *
 * lmEQTL
 *
 * map eQTL by linear model. 
 * 
 * Could take one or two covariate for each marker. 
 * The p-value for the marker is evaluated by F-test.
 *
 * Could carry out cis- only eQTL mapping 
 *
 * Resutls are wrttien into two files. 
 **********************************************************************/

int lmEQTL (int* dims, double* ve, double* vm1, double* vm2, int* ve_na, 
  int* vm1_na, int* vm2_na, char** output, double* RP_cut, int* cis_only, 
  int* cis_distance, int* eChr, int* ePos, int* mChr, int* mPos, 
  double* Rtol, double* Rnna_percent, int* succeed)
{

  int i, j, k, c, pos_diff, n_covariate;
  double chisq0, t_stat, Fstat, Fp, df1, df2, tss;
  int DEBUG = 0;
  
  gsl_matrix * mm2;
  gsl_matrix_int * im2;
  gsl_vector_view md1_v;
  gsl_vector_view md2_v;
  gsl_vector * md2;
  gsl_vector_int * mi2;
  gsl_vector_view gd_a;
  
  /* rank of model matrix */
  size_t rank;
  
  /* 
   * p-value frequencies
   * freqs[100] = #{ pval < P_cut }
   * i = 0:99
   * freqs[i]   = #{ pval < [i/100, (i+1)/100) }
   */
   
  double tmp;
  unsigned long freqs[101];
  for(i=0; i<=100; i++){
     freqs[i] = 0;
  }
  
  const int nrow_e = dims[0];
  const int nrow_m = dims[1];
  const int ncol   = dims[2];
  const int use_m2 = dims[3];
  const double P_cut = RP_cut[0];
  const double tol = Rtol[0];
  const double nna_percent = Rnna_percent[0];
  
  Rprintf("\n------------------------------------------------------\n");
  Rprintf("dims=(%d, %d, %d, %d)\t", nrow_e, nrow_m, ncol, use_m2);
  Rprintf("P_cut=%e\n", P_cut);

  gsl_matrix * me  = gsl_matrix_alloc (nrow_e, ncol); // gene expression matrix
  gsl_matrix * mm1 = gsl_matrix_alloc (nrow_m, ncol); // marker genotype matrix
  gsl_matrix_int * ie  = gsl_matrix_int_alloc (nrow_e, ncol); // missing index
  gsl_matrix_int * im1 = gsl_matrix_int_alloc (nrow_m, ncol); // missing index
  
  n_covariate = 2;
  if(use_m2){
    n_covariate = 3;
    mm2 = gsl_matrix_alloc (nrow_m, ncol); // marker genotype 
    im2 = gsl_matrix_int_alloc (nrow_m, ncol); // missing index
    md2 = gsl_vector_alloc (ncol); // marker genotype data
    mi2 = gsl_vector_int_alloc (ncol); // missing value index
  }
  
  gsl_vector * gd  = gsl_vector_alloc (ncol); // gene expression data
  gsl_vector * md1 = gsl_vector_alloc (ncol); // marker genotype data
  gsl_vector_int * gi  = gsl_vector_int_alloc (ncol); // missing value index
  gsl_vector_int * mi1 = gsl_vector_int_alloc (ncol); // missing value index

  // intercept to be used in model matrix
  gsl_vector * intercept = gsl_vector_alloc (ncol); 
  gsl_vector_set_all(intercept, 1.0);

  // the model matrix
  gsl_matrix * X = gsl_matrix_alloc (ncol, n_covariate);
  gsl_matrix_view Xa;

  // reorganize the data into gsl_vector/matrix structure
  for(j=0; j<ncol; j++){
    for(i=0; i<nrow_e; i++){
      gsl_matrix_set(me, i, j, ve[j*nrow_e+i]);
      gsl_matrix_int_set(ie, i, j, ve_na[j*nrow_e+i]);
    }
  }
  
  for(j=0; j<ncol; j++){
    for(i=0; i<nrow_m; i++){
      gsl_matrix_set(mm1, i, j, vm1[j*nrow_m+i]);
      gsl_matrix_int_set(im1, i, j, vm1_na[j*nrow_m+i]);
      if(use_m2){
        gsl_matrix_set(mm2, i, j, vm2[j*nrow_m+i]);
        gsl_matrix_int_set(im2, i, j, vm2_na[j*nrow_m+i]);
      }
    }
  }
  
  /* 
   * coefficients, p-value of each coefficient and 
   * covariance matrix of the coefficients
   */
  gsl_vector * c0 = gsl_vector_alloc (n_covariate);
  gsl_vector * c0_p = gsl_vector_alloc (n_covariate);
  gsl_matrix * c0_cov = gsl_matrix_alloc (n_covariate, n_covariate);

  // work space for linear model fitting
  gsl_multifit_linear_workspace * work;

  // output file handles
  FILE * fo, *ff;
    
  // time records
  time_t sec_s;
  time_t sec_e;

  sec_s  = time(NULL);
  
  if(DEBUG){
    Rprintf("matrix of me[1, 1:5]\n");
    Rprint_matrix(me, 0, 0, 0, 4, 0);
    Rprintf("matrix of mm1[1, 1:5]\n");
    Rprint_matrix(mm1, 0, 0, 0, 4, 0);
    if(use_m2){
      Rprintf("matrix of mm2[1, 1:5]\n");
      Rprint_matrix(mm2, 0, 0, 0, 4, 0);
    }
  }
  
  fo  = fopen (output[0], "w");
  
  /**
   * write out the header in the output file
   */
  fprintf(fo, "Gene_ID\tMarker_ID\t");
  for(k=0; k<n_covariate; k++){
    fprintf(fo, "b%d\tb%d_se\tb%d_p\t", k, k, k);
  }
  fprintf(fo, "Fstat\tF_p\tN\n");

  /***
   * identifify eQTL gene by gene
   */
  for(i=0; i<nrow_e; i++){ // the (i+1)-th gene expression profile
  
    if(i%1000 == 0){ Rprintf("i=%d\n", i); }
    gsl_vector_view gd_v = gsl_matrix_row (me, i);
    gsl_matrix_int_get_row (gi, ie, i);
    
    // check whether there is enough observations for gi
    c = 0;
    for(k = 0; k<ncol; k++){
      if(gsl_vector_int_get(gi,k) == 1){ c++; }
    }
    if(c < ncol * nna_percent){ continue; }
    
    for(j=0; j<nrow_m; j++){ // the (j+1)-th marker genotype profile
      // Rprintf("  j=%d\n", j); 
      if(*cis_only){
        if(eChr[i] != mChr[j]){
          continue;
        }
        pos_diff = ePos[i] - mPos[j];
        if(pos_diff > *cis_distance || pos_diff < - *cis_distance){
          continue;
        }
      }
      
      md1_v = gsl_matrix_row (mm1, j);
      gsl_matrix_int_get_row (mi1, im1, j);
      
      // after multiplication, mi is the index whether there is 
      // missing value, either gene expression or genotype
      gsl_vector_int_mul (mi1, gi);
      
      if(use_m2){
        md2_v = gsl_matrix_row (mm2, j);
        gsl_matrix_int_get_row (mi2, im2, j);
        gsl_vector_int_mul (mi1, mi2);
      }
      
      c = 0;
      for(k = 0; k<ncol; k++){
        if(gsl_vector_int_get(mi1,k) == 1){
          gsl_vector_set(gd,  c, gsl_vector_get(&gd_v.vector,k));
          gsl_vector_set(md1, c, gsl_vector_get(&md1_v.vector,k));
          if(use_m2)
            gsl_vector_set(md2, c, gsl_vector_get(&md2_v.vector,k));
          c++;
        }
      }
      
      // if there is no enough observation left, exit
      if(c < ncol * nna_percent){ continue; }
                
      // generate the model matrix
      // g = intercept + m.add
      gsl_matrix_set_col(X, 0, intercept);
      gsl_matrix_set_col(X, 1, md1);
      if(use_m2)
        gsl_matrix_set_col(X, 2, md2);
      // Rprint_matrix_i(X, 0, ncol-1, 0, 1, 0);

      // vector include non-missing values
      gd_a = gsl_vector_subvector(gd, 0, c);

      // Rprintf("    finish perparation for linear model\n", j); 

      // model 0: g ~ add
      Xa   = gsl_matrix_submatrix(X, 0, 0, c, n_covariate);
      work = gsl_multifit_linear_alloc (c, n_covariate);
      gsl_multifit_linear_svd (&Xa.matrix, &gd_a.vector, tol, 
                &rank, c0, c0_cov, &chisq0, work);
      gsl_multifit_linear_free(work);

      // Rprintf("    finish linear model\n", j); 

      for(k=0; k<n_covariate; k++){
        t_stat = gsl_vector_get(c0,k)/sqrt(gsl_matrix_get(c0_cov,k,k));
        gsl_vector_set(c0_p,k,2*gsl_cdf_tdist_Q(fabs(t_stat),c-n_covariate));
      }
      
      // Rprintf("    finish t-stat\n", j);
      
      tss   = gsl_stats_tss(gd->data, gd->stride, c);
      df1   = (double) (n_covariate - 1);
      df2   = (double) (c - n_covariate);
      Fstat = (tss - chisq0)*df2/(chisq0*df1);
      
      if (Fstat < 1e-5) { continue; }
      
      Fp    = gsl_cdf_fdist_Q(Fstat, df1, df2);
      k     = (int) (Fp / 0.01);
      
      // Rprintf("tss=%.2e, chisq0=%.2e, Fstat=%.2e", tss, chisq0, Fstat);
      // Rprintf("df1=%d, df2=%d\n", df1, df2)
      freqs[k] += 1;
            
      if(Fp < P_cut){
        freqs[100] += 1;
        fprintf(fo, "%d\t%d\t", i+1, j+1);
        for(k=0; k<n_covariate; k++){
           fprintf(fo, "%.3f\t", gsl_vector_get(c0, k));
           fprintf(fo, "%.3e\t", sqrt(gsl_matrix_get(c0_cov,k,k)));
           fprintf(fo, "%.2e\t", gsl_vector_get(c0_p,k));
        }
        fprintf(fo, "%.3f\t%.2e\t%d\n", Fstat, Fp, c);
      }
    }
  }
  
  fclose(fo);
  
  // print out the frequencies
  ff  = fopen(output[1], "w");
  tmp = 0.0;
  
  fprintf(ff, "<%.2e", P_cut);
  for(i=0; i<100; i++){
    fprintf(ff, "\t%.2f-%.2f", tmp, tmp+0.01);
    tmp += 0.01;
  }
  fprintf(ff, "\n");
  
  fprintf(ff, "%lu", freqs[100]);
  for(i=0; i<100; i++){
    fprintf(ff, "\t%lu", freqs[i]);
  }
  fprintf(ff, "\n");  
  fclose(ff);

  // finish up
  sec_e  = time(NULL);
  Rprintf("total time spent is %ld secs\n", sec_e-sec_s);
  Rprintf("------------------------------------------------------\n");

  gsl_matrix_free(me);
  gsl_matrix_int_free(ie);
  
  gsl_matrix_free(mm1);
  gsl_matrix_int_free(im1);
  
  if(use_m2){
    gsl_matrix_free(mm2);
    gsl_matrix_int_free(im2);
    gsl_vector_free(md2);
    gsl_vector_int_free(mi2);
  }
  
  gsl_vector_free(gd);
  gsl_vector_free(md1);
  gsl_vector_int_free(gi);
  gsl_vector_int_free(mi1);
  gsl_vector_free(c0);
  gsl_vector_free(c0_p);
  gsl_matrix_free(c0_cov);
  gsl_vector_free(intercept);
  gsl_matrix_free(X);
    
  *succeed = 1;
  return(1);
}

/**********************************************************************
 *
 * lmEQTL_max1
 *
 * map eQTL by linear model. 
 * 
 * Could take one or two covariate for each marker. 
 * The p-value for the marker is evaluated by F-test.
 *
 * Could carry out cis- only eQTL mapping 
 *
 * This function is a simplified version of lmEQTL to be used by 
 * function lmEQTL_permute
 *
 * the output is the best (associated) marker oand the corresponding 
 * p-value for each gene
 *
 **********************************************************************/

void lmEQTL_max1 (int* dims, gsl_matrix * me, gsl_matrix * mm1, 
  gsl_matrix * mm2, gsl_matrix_int * ie, gsl_matrix_int * im1, 
  gsl_matrix_int * im2, int* best_m, double* pval, 
  int* cis_only, int* cis_distance, int* eChr, int* ePos, 
  int* mChr, int* mPos, double tol, double nna_percent)
{

  int i, j, k, c, pos_diff, n_covariate, bestM;
  double chisq0, t_stat, Fstat, Fp, Fp_min, df1, df2, tss;
  
  /* rank of model matrix */
  size_t rank;

  const int nrow_e = dims[0];
  const int nrow_m = dims[1];
  const int ncol   = dims[2];
  const int use_m2 = dims[3];

  gsl_vector_view md1_v;
  gsl_vector_view md2_v;
  gsl_vector * md2;
  gsl_vector_int * mi2;
  gsl_vector_view gd_a;

  gsl_vector * gd  = gsl_vector_alloc (ncol); // gene expression data
  gsl_vector * md1 = gsl_vector_alloc (ncol); // marker genotype data
  gsl_vector_int * gi  = gsl_vector_int_alloc (ncol); // missing value index
  gsl_vector_int * mi1 = gsl_vector_int_alloc (ncol); // missing value index
  
  n_covariate = 2;
  if(use_m2){
    n_covariate = 3;
    md2 = gsl_vector_alloc (ncol); // marker genotype data
    mi2 = gsl_vector_int_alloc (ncol); // missing value index
  }

  // intercept to be used in model matrix
  gsl_vector * intercept = gsl_vector_alloc (ncol); 
  gsl_vector_set_all(intercept, 1.0);

  // the model matrix
  gsl_matrix * X = gsl_matrix_alloc (ncol, n_covariate);
  gsl_matrix_view Xa;

  // coefficients different models
  gsl_vector * c0 = gsl_vector_alloc (n_covariate);
  gsl_vector * c0_p = gsl_vector_alloc (n_covariate);
  gsl_matrix * c0_cov = gsl_matrix_alloc (n_covariate, n_covariate);

  // work space for linear model fitting
  gsl_multifit_linear_workspace * work;
  
  // Rprintf("lmEQTL_max1: memory allocated\n");
    
  /***
   * identifify eQTL gene by gene
   */
  for(i=0; i<nrow_e; i++){ // the (i+1)-th gene expression profile
    gsl_vector_view gd_v = gsl_matrix_row (me, i);
    gsl_matrix_int_get_row (gi, ie, i);
    
    // check whether there is enough observations for gi
    c = 0;
    for(k = 0; k<ncol; k++){
      if(gsl_vector_int_get(gi,k) == 1){ c++; }
    }
    if(c < ncol * nna_percent){ continue; }

    Fp_min = 1.0;
    bestM  = -1;
    
    for(j=0; j<nrow_m; j++){ // the (j+1)-th marker genotype profile
      if(*cis_only){
        if(eChr[i] != mChr[j]){
          continue;
        }
        pos_diff = ePos[i] - mPos[j];
        if(pos_diff > *cis_distance || pos_diff < - *cis_distance){
          continue;
        }
      }
      
      md1_v = gsl_matrix_row (mm1, j);
      gsl_matrix_int_get_row (mi1, im1, j);
            
      // after multiplication, mi is the index whether there is 
      // missing value, either gene expression or genotype
      gsl_vector_int_mul (mi1, gi);
      
      if(use_m2){
        md2_v = gsl_matrix_row (mm2, j);
        gsl_matrix_int_get_row (mi2, im2, j);
        gsl_vector_int_mul (mi1, mi2);
      }
      
      c = 0;
      for(k = 0; k<ncol; k++){
        if(gsl_vector_int_get(mi1,k) == 1){
          gsl_vector_set(gd,  c, gsl_vector_get(&gd_v.vector,k));
          gsl_vector_set(md1, c, gsl_vector_get(&md1_v.vector,k));
          if(use_m2)
            gsl_vector_set(md2, c, gsl_vector_get(&md2_v.vector,k));
          c++;
        }
      }

      // if there is not enough observations left, exit
      if(c < ncol * nna_percent){ continue; }
                
      // generate the model matrix
      // g = intercept + m.add
      gsl_matrix_set_col(X, 0, intercept);
      gsl_matrix_set_col(X, 1, md1);
      if(use_m2)
        gsl_matrix_set_col(X, 2, md2);
      // Rprint_matrix_i(X, 0, ncol-1, 0, 1, 0);

      // vector include non-missing values
      gd_a = gsl_vector_subvector(gd, 0, c);

      // model 0: g ~ add
      Xa  = gsl_matrix_submatrix(X, 0, 0, c, n_covariate);
      work = gsl_multifit_linear_alloc (c, n_covariate);
      gsl_multifit_linear_svd (&Xa.matrix, &gd_a.vector, tol, 
                &rank, c0, c0_cov, &chisq0, work);
      gsl_multifit_linear_free(work);

      for(k=0; k<n_covariate; k++){
        t_stat = gsl_vector_get(c0,k)/sqrt(gsl_matrix_get(c0_cov,k,k));
        gsl_vector_set(c0_p,k,2*gsl_cdf_tdist_Q(fabs(t_stat),c-2));
      }
      
      tss   = gsl_stats_tss(gd->data, gd->stride, c);
      df1   = (double) (n_covariate - 1);
      df2   = (double) (c - n_covariate);
      Fstat = (tss - chisq0)*df2/(chisq0*df1);      
      Fp    = gsl_cdf_fdist_Q(Fstat, df1, df2);
      
      if(Fp < Fp_min){
        Fp_min = Fp;
        bestM  = j;
      }
    }
    
    pval[i] = Fp_min;
    best_m[i] = bestM;
  }
  
  gsl_vector_free(gd);
  gsl_vector_free(md1);
  gsl_vector_int_free(gi);
  gsl_vector_int_free(mi1);
  
  if(use_m2){
    gsl_vector_free(md2);
    gsl_vector_int_free(mi2);
  }

  gsl_vector_free(c0);
  gsl_vector_free(c0_p);
  gsl_matrix_free(c0_cov);
  gsl_vector_free(intercept);
  gsl_matrix_free(X);
}

/**********************************************************************
 *
 * lmEQTL_permute
 *
 * map eQTL by linear model. Could take one or two covariate for each
 * marker. The p-value for the marker is evaluated by F-test.
 *
 * Could carry out cis- only eQTL mapping 
 *
 * Carry out fixed number of permutations, and calculate permutation
 * p-value of the best association of each gene expression trait
 * 
 **********************************************************************/

int lmEQTL_permute (int* dims, double* ve, double* vm1, double* vm2, 
  int* ve_na, int* vm1_na, int* vm2_na, int* best_m, 
  double* pval, double* p_pval, int* cis_only, int* cis_distance, 
  int* eChr, int* ePos, int* mChr, int* mPos, int* trace, double* Rtol, 
  double* Rnna_percent, int* Ridum, int* Rgrp, int* Rngrp, int* Rgrpn, 
  int* Rgrps, int* succeed)
{
  int i, j, p, q, *best_m0, *perm1, grp=*Rgrp, ngrp=*Rngrp, **grps;
  double *pval0;
  time_t timer;
  long idum = (long) *Ridum;

  const int nrow_e = dims[0];
  const int nrow_m = dims[1];
  const int ncol   = dims[2];
  const int use_m2 = dims[3];
  const int npermute = dims[4]; // the number of permutations
  
  const double tol   = *Rtol;
  const double nna_percent = *Rnna_percent;

  gsl_matrix * mm2;
  gsl_matrix_int * im2;

  /***
   * best (associated) marker and p-value for each gene 
   * in each permutation
   */  
  best_m0 = (int *)R_alloc(nrow_e, sizeof(int));
  pval0   = (double *)R_alloc(nrow_e, sizeof(double));
  
  /* pernutation index */
  perm1   = (int *)R_alloc(ncol, sizeof(int));

  /* pernutation groups */
  grps    = (int**) R_alloc(ngrp, sizeof(int*));
  grps[0] = (int*) R_alloc(ngrp*ncol, sizeof(int));
  for(i=1; i<ngrp; i++){
  	grps[i] = grps[0] + i*ncol;
  }
  
  gsl_matrix * me  = gsl_matrix_alloc (nrow_e, ncol); // gene expression matrix
  gsl_matrix * mm1 = gsl_matrix_alloc (nrow_m, ncol); // marker genotype matrix
  gsl_matrix_int * ie  = gsl_matrix_int_alloc (nrow_e, ncol); // missing index
  gsl_matrix_int * im1 = gsl_matrix_int_alloc (nrow_m, ncol); // missing index

  if(use_m2){
    mm2 = gsl_matrix_alloc (nrow_m, ncol); // marker genotype 
    im2 = gsl_matrix_int_alloc (nrow_m, ncol); // missing index
  }
  
  // permuted data
  gsl_matrix * pme = gsl_matrix_alloc (nrow_e, ncol); // gene expression matrix
  gsl_matrix_int * pie = gsl_matrix_int_alloc (nrow_e, ncol); // missing index

  // reorganize the data into gsl_vector/matrix structure
  for(j=0; j<ncol; j++){
    for(i=0; i<nrow_e; i++){
      gsl_matrix_set(me, i, j, ve[j*nrow_e+i]);
      gsl_matrix_int_set(ie, i, j, ve_na[j*nrow_e+i]);
    }
  }
  
  for(j=0; j<ncol; j++){
    for(i=0; i<nrow_m; i++){
      gsl_matrix_set(mm1, i, j, vm1[j*nrow_m+i]);
      gsl_matrix_int_set(im1, i, j, vm1_na[j*nrow_m+i]);
      if(use_m2){
        gsl_matrix_set(mm2, i, j, vm2[j*nrow_m+i]);
        gsl_matrix_int_set(im2, i, j, vm2_na[j*nrow_m+i]);
      }
    }
  }
  
  /***
   * finish data delariation and memory allocation
   * start computations
   */

  timer=time(NULL);
  Rprintf("\n------------------------------------------------------\n");
  Rprintf("dims=(%d, %d, %d, %d, %d)\n", nrow_e, nrow_m, ncol,use_m2,npermute);
  
  if(grp){
    Rprintf("permutation by grpup, ngrp=%d\n", ngrp);
    /* reorganize the data into gsl_vector/matrix structure */
    for(j=0; j<ncol; j++){
      for(i=0; i<ngrp; i++){
        grps[i][j] = Rgrps[j*ngrp+i];
      }
    }
  }

  /***
   * find the smallest p-value for each gene with unpermuted data 
   */

  Rprintf("unpermuted data %s\n",asctime(localtime(&timer)));

  lmEQTL_max1 (dims, me, mm1, mm2, ie, im1, im2, best_m, pval, 
    cis_only, cis_distance, eChr, ePos, mChr, mPos, tol, nna_percent);
  
  /***
   * carry out permutations 
   */

  for(p=1; p<=npermute; p++){
    timer=time(NULL);
    
    if(*trace > 0){
      if(p % *trace == 0)
        Rprintf("%d-th permutation %s\n", p, asctime(localtime(&timer)));
    }
    
    if(grp){
    	getPermuteGrp(perm1, ncol, &idum, grps, Rgrpn, ngrp);    	
    }else{
      getPermute(perm1, ncol, &idum);
    }

    permute(me, pme, ie, pie, perm1, nrow_e, ncol);

    lmEQTL_max1 (dims, pme, mm1, mm2, pie, im1, im2, best_m0, pval0, 
      cis_only, cis_distance, eChr, ePos, mChr, mPos, tol, nna_percent);
    
    for(q=0; q<nrow_e; q++){
      if(pval0[q] <= pval[q]){ p_pval[q] += 1.0; }
    }
  }
  
  for(q=0; q<nrow_e; q++){
    p_pval[q] /= npermute;
  }
  
  gsl_matrix_free(me);
  gsl_matrix_free(pme);
  gsl_matrix_free(mm1);
  gsl_matrix_int_free(ie);
  gsl_matrix_int_free(pie);
  gsl_matrix_int_free(im1);

  if(use_m2){
    gsl_matrix_free(mm2);
    gsl_matrix_int_free(im2);
  }
  
  *succeed = 1;
  return(1);
}
